"""Define the core `Entity` class.

Defines the core `Entity` class, which extends `mammos_units.Quantity` to
link physical quantities to ontology concepts. Also includes helper functions
for inferring the correct SI units from the ontology.
"""

from __future__ import annotations

import re
import warnings
from typing import TYPE_CHECKING

import mammos_units as u
from numpy import typing

from mammos_entity._onto import HAVE_INTERNET, mammos_ontology

if TYPE_CHECKING:
    import mammos_units
    import owlready2

    import mammos_entity


base_units = [u.T, u.J, u.m, u.A, u.radian, u.kg, u.s, u.K]


def si_unit_from_list(list_cls: list[owlready2.entity.ThingClass]) -> str:
    """Return an SI unit from a list of entities from the EMMO ontology.

    Given a list of ontology classes, determine which class corresponds to
    a coherent SI derived unit (or if none found, an SI dimensional unit),
    then return that class's UCUM code.

    Args:
        list_cls: A list of ontology classes.

    Returns:
        The UCUM code (e.g., "J/m^3", "A/m") for the first identified SI unit
        in the given list of classes.

    """
    si_unit_cls = [
        cls
        for cls in list_cls
        if mammos_ontology.SICoherentDerivedUnit in cls.ancestors()
    ]
    if not si_unit_cls:
        si_unit_cls = [
            cls
            for cls in list_cls
            if (mammos_ontology.SIDimensionalUnit in cls.ancestors())
        ]
    return si_unit_cls[0].ucumCode[0]


def extract_SI_units(ontology_label: str) -> str | None:
    """Find SI unit for the given label from the EMMO ontology.

    Given a label for an ontology concept, retrieve the corresponding SI unit
    by traversing the class hierarchy. If a valid unit is found, its UCUM code
    is returned; otherwise, None is returned.

    Args:
        ontology_label: The label of an ontology concept
            (e.g., 'SpontaneousMagnetization').

    Returns:
        The UCUM code of the concept's SI unit, or None if no suitable SI unit
        is found or if the unit is a special case like 'Cel.K-1'.

    """
    thing = mammos_ontology.get_by_label(ontology_label)
    si_unit = None
    for ancestor in thing.ancestors():
        if hasattr(ancestor, "hasMeasurementUnit") and ancestor.hasMeasurementUnit:
            if sub_class := list(ancestor.hasMeasurementUnit[0].subclasses()):
                si_unit = si_unit_from_list(sub_class)
            elif ontology_label := ancestor.hasMeasurementUnit[0].ucumCode:
                si_unit = ontology_label[0]
            break
    # HACK: filter Celsius values as Kelvin and `Cel.K-1` as no units
    if si_unit in {"Cel", "mCel"}:
        si_unit = "K"
    elif si_unit == "Cel.K-1":
        si_unit = None
    return si_unit


class Entity(u.Quantity):
    """Create a quantity (a value and a unit) linked to the EMMO ontology.

    Represents a physical property or quantity that is linked to an ontology
    concept. Inherits from `mammos_units.Quantity` and enforces unit
    compatibility with the ontology.

    Args:
        ontology_label: Ontology label
        value: Value
        unit: Unit

    Examples:
        >>> import mammos_entity as me
        >>> m = me.Ms(800000, 'A/m')
        >>> m
        SpontaneousMagnetization(value=800000.0, unit=A / m)

    """

    _repr_latex_ = None

    def __new__(
        cls,
        ontology_label: str,
        value: float | int | typing.ArrayLike = 0,
        unit: str | None = None,
        **kwargs,
    ) -> mammos_units.Quantity:
        if HAVE_INTERNET:
            si_unit = extract_SI_units(ontology_label)
            if (si_unit is not None) and (unit is not None):
                if not u.Unit(si_unit).is_equivalent(unit):
                    raise TypeError(
                        f"The unit {unit} does not match the units of {ontology_label}"
                    )
            elif (si_unit is not None) and (unit is None):
                with u.add_enabled_aliases({"Cel": u.K, "mCel": u.K}):
                    comp_si_unit = u.Unit(si_unit).decompose(bases=base_units)
                unit = u.CompositeUnit(1, comp_si_unit.bases, comp_si_unit.powers)
            elif (si_unit is None) and (unit is not None):
                raise TypeError(
                    f"{ontology_label} is a unitless entity."
                    f" Hence, {unit} is inapropriate."
                )
        else:
            warnings.warn(
                message="Failed to load ontology from the interent"
                ". Hence, no check for unit or ontology_label will be performed!",
                category=RuntimeWarning,
                stacklevel=1,
            )
        comp_unit = u.Unit(unit if unit else "")

        # Remove any set equivalency to enforce unit strictness
        with u.set_enabled_equivalencies([]):
            out = super().__new__(cls, value=value, unit=comp_unit, **kwargs)
        out._ontology_label = ontology_label
        return out

    @property
    def ontology_label(self) -> str:
        """The ontology label that links the entity to the EMMO ontology.

        Retrieve the ontology label corresponding to the `ThingClass` that defines the
        given entity in ontology.

        Returns:
            str: The ontology label corresponding to the right ThingClass.

        """
        return self._ontology_label

    # FIX: right not this will fail if no internet!
    @property
    def ontology(self) -> owlready2.entity.ThingClass:
        """Retrieve the ontology class corresponding to the entity's label.

        Returns:
            The ontology class from `mammos_ontology` that matches the entity's label.

        """
        return mammos_ontology.get_by_label(self.ontology_label)

    @property
    def quantity(self) -> mammos_units.Quantity:
        """Return the entity as a `mammos_units.Quantity`.

        Return a stand-alone `mammos_units.Quantity` object with the same value
        and unit, detached from the ontology link.

        Returns:
            A copy of this entity as a pure physical quantity.

        """
        return u.Quantity(self.value, self.unit)

    @property
    def si(self) -> mammos_entity.Entity:
        """Return the entity in SI units.

        Returns:
            Entity in SI units.

        """
        si_quantity = self.quantity.si
        return self.__class__(
            ontology_label=self.ontology_label,
            value=si_quantity.value,
            unit=si_quantity.unit,
        )

    @property
    def axis_label(self) -> str:
        """Return an ontology based axis label for the plots.

        Returns:
            A string for labelling the axis corresponding to the entity on a plot.

        """
        return (
            re.sub(r"(?<!^)(?=[A-Z])", " ", f"{self.ontology_label}")
            + f" [{self.unit}]"
        )

    def to(self, *args, **kwargs) -> mammos_units.Quantity | mammos_entity.Entity:
        """Modify the unit of the entity in accordance to the EMMO ontology.

        Override method to convert from one unit to the other. If the coversion requires
        equivalencies, the method returns a `astropy.unit.Quantity` otherwise it returns
        an `Entity` with modified units.

        Args:
            unit: The string defining the target unit to convert to (e.g., 'mJ/m').
            equivalencies: List of equivalencies to be used for unit conversion.
            copy: If `True`, then the value is copied.  Otherwise, a copy will only be
                made if necessary.

        Returns:
            mammos_units.Quantity if equivalencies are used to convert the units,
            mammos_entity.Entity if equivalencies are not used to convert the units.

        """
        quantity = self.quantity.to(*args, **kwargs)
        with u.set_enabled_equivalencies([]):
            if self.quantity.unit.is_equivalent(quantity.unit):
                return self.__class__(
                    ontology_label=self.ontology_label, value=quantity
                )
            else:
                return quantity

    def __repr__(self) -> str:
        new_line = "\n" if self.value.size > 4 else ""
        if self.unit.is_equivalent(u.dimensionless_unscaled):
            repr_str = f"{self.ontology_label}(value={new_line}{self.value})"
        else:
            repr_str = (
                f"{self.ontology_label}(value={new_line}{self.value}"
                f",{new_line} unit={self.unit})"
            )
        return repr_str

    def __str__(self) -> str:
        return self.__repr__()

    def __array_ufunc__(self, func, method, *inputs, **kwargs):
        """Override NumPy universal functions in case of mathematical operations.

        Override NumPy's universal functions to return a regular quantity rather
        than another `Entity` when performing array operations (e.g., add, multiply)
        since these oprations change the units.
        """
        result = super().__array_ufunc__(func, method, *inputs, **kwargs)

        if isinstance(result, self.__class__):
            return result.quantity
        else:
            return result



"""Entity (Quantity and EMMO ontology label).

Exposes the primary components of the MaMMoS entity package, including
the `Entity` class for ontology-linked physical quantities, pre-defined
factory methods for common magnetic entities (Ms, A, Ku, H), and the
loaded MaMMoS ontology object.
"""

import importlib.metadata

from mammos_entity._base import Entity
from mammos_entity._entities import A, B, BHmax, H, Hc, Ku, M, Mr, Ms, T, Tc
from mammos_entity._onto import mammos_ontology

__version__ = importlib.metadata.version(__package__)


__all__ = [
    "Entity",
    "A",
    "B",
    "BHmax",
    "H",
    "Hc",
    "Ku",
    "M",
    "Mr",
    "Ms",
    "T",
    "Tc",
    "mammos_ontology",
]



"""Convenience factory functions for often used entities.

Provides convenience factory functions for creating common physical property
entities (such as spontaneous magnetization or external magnetic field)
using the `Entity` class from `mammos_entity.base`.
"""

from __future__ import annotations

from typing import TYPE_CHECKING

from numpy import typing

from mammos_entity._base import Entity

if TYPE_CHECKING:
    import mammos_entity


def Ms(
    value: int | float | typing.ArrayLike = 0, unit: None | str = None
) -> mammos_entity.Entity:
    """Create an Entity representing the spontaneous magnetization (Ms).

    Args:
        value: Numeric value corresponding to spontaneous magnetization. It can also be
            a Numpy array.
        unit: Unit of measure for the value (e.g., 'A/m'). If omitted, the SI unit
            from the ontology, i.e. A/m, will be inferred.

    Returns:
        An `Entity` object labelled "SpontaneousMagnetization".

    """
    return Entity("SpontaneousMagnetization", value, unit)


def A(
    value: int | float | typing.ArrayLike = 0, unit: None | str = None
) -> mammos_entity.Entity:
    """Create an Entity representing the exchange stiffness constant (A).

    Args:
        value: Numeric value corresponding to exchange stiffness. It can also be a
            Numpy array.
        unit: Unit of measure for the value (e.g., 'J/m'). If omitted, the SI unit
            from the ontology, i.e. J/m, will be inferred.

    Returns:
        An `Entity` object labeled "ExchangeStiffnessConstant".

    """
    return Entity("ExchangeStiffnessConstant", value, unit)


def Ku(
    value: int | float | typing.ArrayLike = 0, unit: None | str = None
) -> mammos_entity.Entity:
    """Create an Entity representing the uniaxial anisotropy constant (Ku).

    Args:
        value: Numeric value corresponding to the uniaxial anisotropy constant. It can
            also be a Numpy array.
        unit: Unit of measure for the value (e.g., 'J/m^3'). If omitted, the SI unit
            from the ontology, i.e. J/m^3 will be inferred.

    Returns:
        An `Entity` object labeled "UniaxialAnisotropyConstant".

    """
    return Entity("UniaxialAnisotropyConstant", value, unit)


def H(
    value: int | float | typing.ArrayLike = 0, unit: None | str = None
) -> mammos_entity.Entity:
    """Create an Entity representing the external magnetic field (H).

    Args:
        value: Numeric value corresponding to the external magnetic field. It can also
            be a Numpy array.
        unit: Unit of measure for the value (e.g., 'T' for Tesla). If omitted, the SI
            unit from the ontology, i.e. T, will be inferred.

    Returns:
        Entity: An `Entity` object labeled "ExternalMagneticField".


    """
    return Entity("ExternalMagneticField", value, unit)


def Tc(
    value: int | float | typing.ArrayLike = 0, unit: None | str = None
) -> mammos_entity.Entity:
    """Create an Entity representing the Curie temperature (Tc).

    Args:
        value: Numeric value corresponding to the Curie temperature. It can also be a
            Numpy array.
        unit : Unit of measure for the value (e.g., 'K' for Kelvin). If omitted, the SI
            unit from the ontology, i.e. K, will be inferred.

    Returns:
        An `Entity` object labelled "CurieTemperature".

    """
    return Entity("CurieTemperature", value, unit)


def Hc(
    value: int | float | typing.ArrayLike = 0, unit: None | str = None
) -> mammos_entity.Entity:
    """Create an Entity representing the external coercive field (Hc).

    Args:
        value : Numeric value corresponding to the external coercive field. It can also
            be a Numpy array.
        unit : Unit of measure for the value (e.g., 'A/m'). If omitted, the SI unit
            from the ontology, i.e. A/m, will be inferred.

    Returns:
        Entity: An `Entity` object labelled "CoercivityHcExternal".

    """
    return Entity("CoercivityHcExternal", value, unit)


def Mr(
    value: int | float | typing.ArrayLike = 0, unit: None | str = None
) -> mammos_entity.Entity:
    """Create an Entity representing the remanent magnetisation (Mr).

    Args:
        value : Numeric value corresponding to the remanent magnetisation. It can also
            be a Numpy array.
        unit : Unit of measure for the value (e.g., 'A/m'). If omitted, the SI unit
            from the ontology, i.e. A/m, will be inferred.

    Returns:
        An `Entity` object labelled "Remanence".

    """
    return Entity("Remanence", value, unit)


def BHmax(
    value: int | float | typing.ArrayLike = 0, unit: None | str = None
) -> mammos_entity.Entity:
    """Create an Entity representing the maximum energy product of the hysteresis loop.

    Args:
        value : Numeric value corresponding to the maximum energy product. It can also
            be a Numpy array.
        unit : Unit of measure for the value (e.g., 'J/m3'). If omitted, the SI unit
            from the ontology, i.e. J/m3, will be inferred.

    Returns:
        An `Entity` object labelled "MaximumEnergyProduct".

    """
    return Entity("MaximumEnergyProduct", value, unit)


def T(
    value: int | float | typing.ArrayLike = 0, unit: None | str = None
) -> mammos_entity.Entity:
    """Create an Entity representing the temperature (T).

    Args:
        value : Numeric value corresponding to the temperature. It can also
            be a Numpy array.
        unit : Unit of measure for the value (e.g., 'K'). If omitted, the SI unit
            from the ontology, i.e. K, will be inferred.

    Returns:
        An `Entity` object labelled "ThermodynamicTemperature".

    """
    return Entity("ThermodynamicTemperature", value, unit)


def B(
    value: int | float | typing.ArrayLike = 0, unit: None | str = None
) -> mammos_entity.Entity:
    """Create an Entity representing the magnetic flux density (B).

    Args:
        value : Numeric value corresponding to the magnetic flux density. It can also
            be a Numpy array.
        unit : Unit of measure for the value (e.g., 'T'). If omitted, the SI unit
            from the ontology, i.e. T, will be inferred.

    Returns:
        An `Entity` object labelled "MagneticFluxDensity".

    """
    return Entity("MagneticFluxDensity", value, unit)


def M(
    value: int | float | typing.ArrayLike = 0, unit: None | str = None
) -> mammos_entity.Entity:
    """Create an Entity representing the magnetization.

    Args:
        value : Numeric value corresponding to the magnetization of the material. It can
            also be a Numpy array.
        unit : Unit of measure for the value (e.g., 'A/m'). If omitted, the SI unit
            from the ontology, i.e. A/m, will be inferred.

    Returns:
        An `Entity` object labelled "Magnetization".

    """
    return Entity("Magnetization", value, unit)



import mammos_units as u
import numpy as np
import pytest

import mammos_entity as me


def test_unit_conversion():
    e = me.A(42)  # NOTE: we know that unit by default J/m
    e_same = me.A(42e3, unit="mJ/m")
    assert np.allclose(e, e_same)


def test_SI_conversion():
    e = me.BHmax(42, unit="kJ/m3")
    e_si = e.si
    assert e.ontology_label == e_si.ontology_label
    assert np.allclose(e, e_si)
    assert e_si.unit == "J/m3"


def test_to_method():
    e = me.H(8e5)
    e_same = e.to("mA/m")
    np.allclose(e, e_same)
    assert e.ontology_label == e_same.ontology_label
    e_eq = e.to("T", equivalencies=u.magnetic_flux_field())
    assert not hasattr(e_eq, "ontology_label")
    assert not isinstance(e_eq, me.Entity)
    assert isinstance(e_eq, u.Quantity)


def test_numpy_array_as_value():
    val = np.array([42, 42, 42])
    e = me.H(val)
    assert np.allclose(e.value, val)


def test_multidim_numpy_array_as_value():
    val = np.ones((42, 42, 42, 3))
    e = me.H(val)
    assert np.allclose(e.value, val)


def test_list_as_value():
    val = [42, 42, 42]
    e = me.Ku(val)
    assert np.allclose(e.value, np.array(val))


def test_tuple_as_value():
    val = (42, 42, 42)
    e = me.Ms(val)
    assert np.allclose(e.value, np.array(val))


def test_entity_drop_ontology_numpy(onto_class_list):
    for label in onto_class_list:
        e = me.Entity(label, 42)
        root_e = np.sqrt(e)
        with pytest.raises(AttributeError):
            _ = root_e.ontology


def test_entity_drop_ontology_multiply(onto_class_list):
    for label in onto_class_list:
        e = me.Entity(label, 42)
        mul_e = e * e
        with pytest.raises(AttributeError):
            _ = mul_e.ontology


def test_all_labels_ontology(onto_class_list):
    for label in onto_class_list:
        _ = me.Entity(label, 42)


def test_quantity_as_value():
    val = 1 * u.A / u.m
    e = me.Ms(val)
    assert u.allclose(e.quantity, val)


def test_wrong_quantity_as_value():
    val = 1 * u.T
    with pytest.raises(u.UnitConversionError):
        me.Ms(val)


def test_quantity_as_value_and_unit():
    val = 1 * u.A / u.m
    e = me.Ms(val, "A/m")
    assert u.allclose(e.quantity, val)


def test_wrong_quantity_as_value_and_unit():
    val = 1 * u.T
    with pytest.raises(u.UnitConversionError):
        me.Ms(val, "A/m")


def test_wrong_quantity_as_value_and_wrong_unit():
    val = 1 * u.T
    with pytest.raises(TypeError):
        me.Ms(val, "T")


def test_wrong_quantity_with_equivalency():
    val = 1 * u.T
    with (
        u.set_enabled_equivalencies(u.magnetic_flux_field()),
        pytest.raises(u.UnitConversionError),
    ):
        me.Ms(val)


def test_wrong_quantity_with_equivalency_early_conversion():
    val = 1 * u.T
    with u.set_enabled_equivalencies(u.magnetic_flux_field()):
        e = me.Ms(val.to(u.A / u.m))
        assert u.allclose(e.quantity, val)



import mammos_units as u
import pytest

import mammos_entity as me


def test_Ms_unit_val():
    e = me.Ms(42)
    assert e.unit == (u.A / u.m)


def test_Ms_unit_allowed():
    allowed_units = ["A/m", "mA/m", "kA/m", "nA/m", "MA/m"]
    for unit in allowed_units:
        _ = me.Ms(42, unit)


def test_Ms_unit_not_allowed():
    unallowed_units = ["T", "A", "J", "m"]
    for unit in unallowed_units:
        with pytest.raises(TypeError):
            _ = me.Ms(42, unit)


def test_Ms_ontology():
    e = me.Ms(42)
    assert str(e.ontology.prefLabel[0]) == "SpontaneousMagnetization"


def test_A_unit_val():
    e = me.A(42)
    assert e.unit == (u.J / u.m)


def test_A_unit_allowed():
    allowed_units = ["J/m", "mJ/m", "kJ/m", "nJ/m", "MJ/m"]
    for unit in allowed_units:
        _ = me.A(42, unit)


def test_A_unit_not_allowed():
    unallowed_units = ["T", "A", "J", "m"]
    for unit in unallowed_units:
        with pytest.raises(TypeError):
            _ = me.A(42, unit)


def test_A_ontology():
    e = me.A(42)
    assert str(e.ontology.prefLabel[0]) == "ExchangeStiffnessConstant"


def test_Ku_unit_val():
    e = me.Ku(42)
    assert e.unit == (u.J / u.m**3)


def test_Ku_unit_allowed():
    allowed_units = ["J/m3", "mJ/m3", "kJ/m3", "nJ/m3", "MJ/m3"]
    for unit in allowed_units:
        _ = me.Ku(42, unit)


def test_Ku_unit_not_allowed():
    unallowed_units = ["T", "A", "J/m2", "m"]
    for unit in unallowed_units:
        with pytest.raises(TypeError):
            _ = me.Ku(42, unit)


def test_Ku_ontology():
    e = me.Ku(42)
    assert str(e.ontology.prefLabel[0]) == "UniaxialAnisotropyConstant"


def test_H_unit_val():
    e = me.H(42)
    assert e.unit == (u.A / u.m)


def test_H_unit_allowed():
    allowed_units = ["A/m", "mA/m", "kA/m", "nA/m", "MA/m"]
    for unit in allowed_units:
        _ = me.H(42, unit)


def test_H_unit_not_allowed():
    unallowed_units = ["T", "A", "J", "m"]
    for unit in unallowed_units:
        with pytest.raises(TypeError):
            _ = me.H(42, unit)


def test_H_ontology():
    e = me.H(42)
    assert str(e.ontology.prefLabel[0]) == "ExternalMagneticField"


def test_Tc_unit_val():
    e = me.Tc(42)
    assert e.unit == u.K


def test_Tc_unit_allowed():
    allowed_units = ["K", "mK", "kK", "nK", "MK"]
    for unit in allowed_units:
        _ = me.Tc(42, unit)


def test_Tc_unit_not_allowed():
    unallowed_units = ["T", "A", "J", "m", "deg_C"]
    for unit in unallowed_units:
        with pytest.raises(TypeError):
            _ = me.Tc(42, unit)


def test_Tc_ontology():
    e = me.Tc(42)
    assert str(e.ontology.prefLabel[0]) == "CurieTemperature"


def test_Hc_unit_val():
    e = me.Hc(42)
    assert e.unit == (u.A / u.m)


def test_Hc_unit_allowed():
    allowed_units = ["A/m", "mA/m", "kA/m", "nA/m", "MA/m"]
    for unit in allowed_units:
        _ = me.Hc(42, unit)


def test_Hc_unit_not_allowed():
    unallowed_units = ["T", "A", "J", "m"]
    for unit in unallowed_units:
        with pytest.raises(TypeError):
            _ = me.Hc(42, unit)


def test_Hc_ontology():
    e = me.Hc(42)
    assert str(e.ontology.prefLabel[0]) == "CoercivityHcExternal"


def test_Mr_unit_val():
    e = me.Mr(42)
    assert e.unit == (u.A / u.m)


def test_Mr_unit_allowed():
    allowed_units = ["A/m", "mA/m", "kA/m", "nA/m", "MA/m"]
    for unit in allowed_units:
        _ = me.Mr(42, unit)


def test_Mr_unit_not_allowed():
    unallowed_units = ["T", "A", "J", "m"]
    for unit in unallowed_units:
        with pytest.raises(TypeError):
            _ = me.Mr(42, unit)


def test_Mr_ontology():
    e = me.Mr(42)
    assert str(e.ontology.prefLabel[0]) == "Remanence"


def test_BHmax_unit_not_allowedval():
    e = me.BHmax(42)
    assert e.unit == (u.J / u.m**3)


def test_BHmax_unit_allowed():
    allowed_units = ["J/m3", "mJ/m3", "kJ/m3", "nJ/m3", "MJ/m3"]
    for unit in allowed_units:
        _ = me.BHmax(42, unit)


def test_BHmax_unit_not_allowed():
    unallowed_units = ["T", "A", "J", "m"]
    for unit in unallowed_units:
        with pytest.raises(TypeError):
            _ = me.BHmax(42, unit)


def test_BHmax_ontology():
    e = me.BHmax(42)
    assert str(e.ontology.prefLabel[0]) == "MaximumEnergyProduct"


def test_unique_labels():
    assert (
        len(
            {
                me.A().ontology_label,
                me.Ms().ontology_label,
                me.Ku().ontology_label,
                me.H().ontology_label,
                me.Hc().ontology_label,
                me.Tc().ontology_label,
                me.Mr().ontology_label,
                me.BHmax().ontology_label,
            }
        )
        == 8
    )



"""Loads and provides access to MaMMoS ontology which is part of EMMO.

Loads and provides access to the MaMMoS magnetic materials ontology via the
`EMMOntoPy` library. The ontology is loaded from a remote TTL (Turtle) file
containing definitions of relevant magnetic material concepts.
"""

import warnings

from ontopy import ontology, utils

HAVE_INTERNET = True

try:
    mammos_ontology = ontology.get_ontology(
        "https://raw.githubusercontent.com/MaMMoS-project/MagneticMaterialsOntology/refs/heads/main/magnetic_material_mammos.ttl"
    ).load()
except utils.EMMOntoPyException:
    warnings.warn(
        message="Failed to load ontology from the internet.",
        category=RuntimeWarning,
        stacklevel=2,
    )
    mammos_ontology = None
    HAVE_INTERNET = False
